﻿using OpenCvSharp;
using SharpDX;
using SharpDX.Direct2D1;
using System;
using System.Collections.Concurrent;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace WindowsFormsApp1
{
    public partial class Form1 : Form
    {
        EventWaitHandle _exitEvent = new EventWaitHandle(false, EventResetMode.ManualReset);
        EventWaitHandle _captureEvent = new EventWaitHandle(false, EventResetMode.AutoReset);
        EventWaitHandle[] _waitSignals;
        Thread _captureThread;
        Thread _fpsThread;

        RenderTarget2D _renderTarget;
        CaptureSource _captureSource;

        [DllImport("OpenCvSharpExtern.dll", EntryPoint = "??0Mat@cv@@QEAA@HHHPEAX_K@Z")]
        internal static unsafe extern void CreateMatWithPtr(IntPtr thisPtr, int rows, int cols, int type, void* data, int step = 0);

        BlockingCollection<SharpDX.Direct2D1.Bitmap> _queue = new BlockingCollection<SharpDX.Direct2D1.Bitmap>();

        public Form1()
        {
            InitializeComponent();

            _captureSource = CaptureSource.Monitor1;
            _waitSignals = new[] { _exitEvent, _captureEvent };

            _renderTarget = new RenderTarget2D();

            _captureThread = new Thread(captureThreadFunc);
            _captureThread.IsBackground = true;
            _captureThread.Start(_viewBox.Handle);

            _fpsThread = new Thread(fpsThreadFunc);
            _fpsThread.IsBackground = true;
            _fpsThread.Start();

        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            _exitEvent.Set();

            _renderTarget.Dispose();

            base.OnFormClosing(e);
        }

        protected unsafe override void OnKeyPress(KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Cancel) // Ctrl + C
            {
                _captureEvent.Set();
            }

            base.OnKeyPress(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            while (_queue.Count != 0)
            {
                if (_queue.TryTake(out SharpDX.Direct2D1.Bitmap bitmap) == true)
                {
                    if (bitmap == null)
                    {
                        return;
                    }

                    _renderTarget.Render(
                        (renderer) =>
                        {
                            renderer.DrawBitmap(bitmap, 1.0f, BitmapInterpolationMode.Linear);
                        });

                    bitmap.Dispose();
                }
            }
        }

        private void captureThreadFunc(object arg)
        {
            DXGIManager manager = null;

            while (true)
            {
                using (manager = new DXGIManager(_captureSource))
                {
                    if (manager.Initialized == false)
                    {
                        Thread.Sleep(500);
                        continue;
                    }

                    this.BeginInvoke((Action)(() =>
                    {
                        this._viewBox.Top = 0;
                        this._viewBox.Left = 0;
                        this._viewBox.Width = manager.Width;
                        this._viewBox.Height = manager.Height;
                    }));

                    _renderTarget.Initialize((IntPtr)arg, manager.Width, manager.Height);

                    // Initialize 호출 후 너무 빠르게 Capture를 호출하면 (검은 색의) 빈 화면이 나올 수 있음.
                    // 필요하다면 Sleep을 호출
                    // Thread.Sleep(500);

                    try
                    {
                        CaptureLoop(manager);
                        break;
                    }
                    catch (SharpDXException e)
                    {
                        if (e.ResultCode != DXGIError.DXGI_ERROR_ACCESS_LOST)
                        {
                            // 예를 들어, Ctrl + Shift + Alt 키를 눌러 데스크탑 전환을 한 경우
                            manager.Dispose();
                        }
                    }
                }
            }
        }

        unsafe void copyFrameBuffer(IntPtr srcPtr, IntPtr dstPtr, int srcPitch, Rectangle offsetBounds)
        {
            //Mat mat = new Mat();
            //byte* pBuf = (byte *)srcPtr.ToPointer();
            //CreateMatWithPtr(mat.CvPtr, offsetBounds.Height, offsetBounds.Width, MatType.CV_8UC4, pBuf);

            Mat mat = new Mat(new int[] { offsetBounds.Height, offsetBounds.Width }, MatType.CV_8UC4, srcPtr);

            using (Mat gray = mat.CvtColor(ColorConversionCodes.BGRA2GRAY))
            using (Mat blur = gray.GaussianBlur(new OpenCvSharp.Size(7, 7), 1.5, 1.5))
            using (Mat canny = blur.Canny(0, 30, 3))
            {
                IntPtr srcMatPtr = canny.Data;

                for (int y = 0; y < _renderTarget.Height; y++)
                {
                    Utilities.CopyMemory(dstPtr + (offsetBounds.Left), srcMatPtr, _renderTarget.Width);

                    srcMatPtr = IntPtr.Add(srcMatPtr, _renderTarget.Width);
                    dstPtr = IntPtr.Add(dstPtr, _renderTarget.Width);
                }
            }
        }

        private void CaptureLoop(DXGIManager manager)
        {
            using (Mat mat = new Mat(new OpenCvSharp.Size(manager.Width, manager.Height), MatType.CV_8UC1))
            {
                while (true)
                {
                    int signalId = EventWaitHandle.WaitAny(_waitSignals, Timeout.Infinite);

                    if (signalId == 0)
                    {
                        break;
                    }

                    IntPtr dstPtr = mat.Data;

                    if (manager.Capture(copyFrameBuffer, dstPtr, 1000) == true)
                    {
                        using (Mat last = mat.CvtColor(ColorConversionCodes.GRAY2BGRA))
                        {
                            DataPointer dataPointer = new DataPointer(last.Data, (int)last.Total() * last.Channels());
                            SharpDX.Direct2D1.Bitmap bitmap = _renderTarget.CreateBitmap(dataPointer);

                            if (bitmap != null)
                            {
                                _queue.Add(bitmap);
                                _count++;

                                try
                                {
                                    this.Invoke((Action)(() => this.Invalidate()));
                                }
                                catch (ObjectDisposedException) { }
                                catch (InvalidOperationException) { }
                            }
                        }
                    }
                }
            }
        }

        long _count;
        private void fpsThreadFunc()
        {
            while (true)
            {
                long oldCount = _count;
                Thread.Sleep(1000);
                long fps = _count - oldCount;

                this.BeginInvoke((Action)(() => this.Text = fps.ToString()));
            }
        }
    }
}
